/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2016                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2015                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef __EXTENDED_IF_EVALUATION_H
#define __EXTENDED_IF_EVALUATION_H 0

#include <vector>
#include <stack>

#include "CubeNullaryEvaluation.h"

namespace cube
{
class ExtendedIfEvaluation : public NullaryEvaluation
{
protected:
    std::vector<std::vector<GeneralEvaluation*> > statement_blocks;
    std::vector<GeneralEvaluation*>               statement_conditions;
public:

    ExtendedIfEvaluation( std::stack<GeneralEvaluation*>              _statement_conditions,
                          std::stack<std::stack<GeneralEvaluation*> > _statement_blocks );


    ~ExtendedIfEvaluation();


    inline
    virtual
    void
    set_verbose_execution( bool _v )
    {
	GeneralEvaluation::set_verbose_execution(_v);
	for ( std::vector<GeneralEvaluation*>::iterator iter = statement_conditions.begin(); iter != statement_conditions.end(); ++iter )
        {
            GeneralEvaluation* arg = *iter;
	    arg->set_verbose_execution( _v );
        }
        for ( std::vector<std::vector<GeneralEvaluation*> >::iterator blocks_iter = statement_blocks.begin(); blocks_iter != statement_blocks.end(); ++blocks_iter )
        {
	    std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
	    for ( std::vector<GeneralEvaluation*>::iterator iter = _arg.begin(); iter != _arg.end(); ++iter )
	    {
		GeneralEvaluation* arg = *iter;
		arg->set_verbose_execution( _v );
	    }
	}
    }
    

    inline
    virtual
    void
    setRowSize( size_t size );

    virtual
    double
    eval();

    virtual
    double
    eval( Cnode* cnode, CalculationFlavour cf, Sysres* sr, CalculationFlavour tf  )
    {
        std::vector<std::vector<GeneralEvaluation*> >::iterator blocks_iter = statement_blocks.begin();
        for ( std::vector<GeneralEvaluation*>::iterator iter = statement_conditions.begin(); iter != statement_conditions.end(); ++iter )
        {
            GeneralEvaluation* arg = *iter;
            if ( arg->eval( cnode, cf, sr, tf ) != 0. )
            {
                std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
                for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
                {
                    GeneralEvaluation* __arg = *_iter;
                    __arg->eval( cnode, cf, sr, tf );
                }
                return 0.;
            }
            ++blocks_iter;
        }

        // at this moment the else is executed
        if ( blocks_iter != statement_blocks.end() )
        {
            std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
            for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
            {
                GeneralEvaluation* __arg = *_iter;
                __arg->eval( cnode, cf, sr, tf );
            }
        }
        return 0.;
    };


    virtual
    double
    eval( Cnode* cnode, CalculationFlavour cf )
    {
        std::vector<std::vector<GeneralEvaluation*> >::iterator blocks_iter = statement_blocks.begin();
        for ( std::vector<GeneralEvaluation*>::iterator iter = statement_conditions.begin(); iter != statement_conditions.end(); ++iter )
        {
            GeneralEvaluation* arg = *iter;
            if ( arg->eval( cnode, cf ) != 0. )
            {
                std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
                for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
                {
                    GeneralEvaluation* __arg = *_iter;
                    __arg->eval( cnode, cf );
                }
                return 0.;
            }
            ++blocks_iter;
        }
        // at this moment the else is executed
        if ( blocks_iter != statement_blocks.end() )
        {
            std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
            for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
            {
                GeneralEvaluation* __arg = *_iter;
                __arg->eval( cnode, cf );
            }
        }

        return 0.;
    };


    virtual
    double
    eval( double arg1, double arg2 )
    {
        std::vector<std::vector<GeneralEvaluation*> >::iterator blocks_iter = statement_blocks.begin();
        for ( std::vector<GeneralEvaluation*>::iterator iter = statement_conditions.begin(); iter != statement_conditions.end(); ++iter )
        {
            GeneralEvaluation* arg = *iter;
            if ( arg->eval(  arg1, arg2 ) != 0. )
            {
                std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
                for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
                {
                    GeneralEvaluation* __arg = *_iter;
                    __arg->eval(  arg1, arg2 );
                }
                return 0.;
            }
            ++blocks_iter;
        }
        // at this moment the else is executed
        if ( blocks_iter != statement_blocks.end() )
        {
            std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
            for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
            {
                GeneralEvaluation* __arg = *_iter;
                __arg->eval(  arg1, arg2 );
            }
        }

        return 0.;
    };

    virtual
    double*
    eval_row( Cnode*             cnode,
              CalculationFlavour cf )
    {
        std::vector<std::vector<GeneralEvaluation*> >::iterator blocks_iter = statement_blocks.begin();
        for ( std::vector<GeneralEvaluation*>::iterator iter = statement_conditions.begin(); iter != statement_conditions.end(); ++iter )
        {
            GeneralEvaluation* arg = *iter;
            if ( arg->eval(  cnode, cf ) != 0. )
            {
                std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
                for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
                {
                    GeneralEvaluation* __arg = *_iter;
                    delete[] ( __arg->eval_row(  cnode, cf ) );
                }
                return NULL;
            }
            ++blocks_iter;
        }
        // at this moment the else is executed
        if ( blocks_iter != statement_blocks.end() )
        {
            std::vector<GeneralEvaluation*>& _arg = *blocks_iter;
            for ( std::vector<GeneralEvaluation*>::iterator _iter = _arg.begin(); _iter != _arg.end(); ++_iter )
            {
                GeneralEvaluation* __arg = *_iter;
                delete[] ( __arg->eval_row(  cnode, cf ) );
            }
        }
        return NULL;
    };


    virtual
    void
    print();
};
};

#endif
